// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests channel performance/scalability in different setups.
// chan-ring tests channel scalability in single-producer/single-consumer mode.
//	A set of goroutines is arranged into "a ring" connected by channels,
//	a number of small messages is passed through the ring.
//	There are 4 variations: unbuf/buf10/buf100/buf1000 that relate to
//	unbuffered channels and buffered channels of size 10/100/1000, respectively.
//	Buffered variations are expected to scale rather good.
// chan-pc tests channel scalability in multi-producer/multi-consumer mode.
//	N producer goroutines and N consumer goroutines stress single channel.
//	There are 4 variations: unbuf/buf10/buf100/buf1000 that relate to
//	unbuffered channels and buffered channels of size 10/100/1000, respectively.
//	work variations execute some amount of local work before/after item production/consumption.
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
)

func benchmarkRing(chansize int) {
	procs := runtime.GOMAXPROCS(-1)
	if procs == 1 && chansize == 0 {
		chansize = 1
	}
	source := make(chan int, chansize)
	sink := source
	if procs != 1 {
		sink = make(chan int, chansize)
	}
	compl := make(chan int, 1)
	go func() {
		sent := uint64(0)
		for benchmarking {
			select {
			case source <- 1:
				sent += 1
				if (sent % 1000) == 0 {
					runtime.Gosched()
				}
			case <-sink:
			}
		}
		stop := false
		for !stop {
			select {
			case <-sink:
			case source <- 0:
				stop = true
			}
		}
		stop = false
		for !stop {
			select {
			case v := <-sink:
				if v == 0 {
					stop = true
				}
			}
		}
		totalWork = sent * uint64(procs)
		compl <- 1
	}()
	prevc := source
	nextc := sink
	for p := 1; p != procs; p += 1 {
		if p != procs-1 {
			nextc = make(chan int, chansize)
		} else {
			nextc = sink
		}
		go func(inc chan int, outc chan int) {
			for true {
				v := <-inc
				outc <- v
				if v == 0 {
					break
				}
			}
		}(prevc, nextc)
		prevc = nextc
	}
	<-compl
}

func benchmarkPC(chansize, localWork int) {
	procs := runtime.GOMAXPROCS(-1) * 2
	queue := make(chan int, chansize)
	compl := make(chan uint64, procs)
	for p := 0; p != procs/2; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				for i := 0; i != localWork; i += 1 {
					work *= 2
					work /= 2
				}
				queue <- 1
			}
			queue <- 0
			compl <- work
		}()
	}
	for p := 0; p != procs/2; p += 1 {
		go func() {
			work := uint64(0)
			for true {
				m := <-queue
				work += 1
				for i := 0; i != localWork; i += 1 {
					work *= 2
					work /= 2
				}
				if m == 0 {
					break
				}
			}
			compl <- work
		}()
	}
	for i := 0; i != procs; i += 1 {
		totalWork += <-compl
	}
}

func SuiteChan(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"chan-ring-unbuf", func() {
		benchmarkRing(0)
	}})
	benchmarks.Push(&Benchmark{"chan-ring-buf10", func() {
		benchmarkRing(10)
	}})
	benchmarks.Push(&Benchmark{"chan-ring-buf100", func() {
		benchmarkRing(100)
	}})
	benchmarks.Push(&Benchmark{"chan-ring-buf1000", func() {
		benchmarkRing(1000)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-unbuf", func() {
		benchmarkPC(0, 0)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf10", func() {
		benchmarkPC(10, 0)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf100", func() {
		benchmarkPC(100, 0)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf1000", func() {
		benchmarkPC(1000, 0)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-unbuf-work", func() {
		benchmarkPC(0, 100)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf10-work", func() {
		benchmarkPC(10, 100)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf100-work", func() {
		benchmarkPC(100, 100)
	}})
	benchmarks.Push(&Benchmark{"chan-pc-buf1000-work", func() {
		benchmarkPC(1000, 100)
	}})
}
