// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests sync.Cond performance/scalability.
// condvar-distr tests scalability of the condition variable without contention.
//	Expected to scale almost linearly.
// condvar-nonblock tests scalability of the condition variable under contention
//	but without actual blocking.
// condvar-block tests scalability of the condition variable under contention
//	and with actual blocking/unblocking.
//	slack variation creates more goroutines than there are worker threads.
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
	"sync"
)

func benchmarkCondvarDistr() {
	procs := runtime.GOMAXPROCS(-1)
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			mtx := &sync.Mutex{}
			cv := sync.NewCond(mtx)
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				cv.Signal()
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func benchmarkCondvarNonblock() {
	procs := runtime.GOMAXPROCS(-1)
	mtx := &sync.Mutex{}
	cv := sync.NewCond(mtx)
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				cv.Signal()
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func benchmarkCondvarBlock(slack int) {
	procs := slack * runtime.GOMAXPROCS(-1)
	if procs%2 != 0 {
		procs += 1
	}
	mtx := &sync.Mutex{}
	cv := sync.NewCond(mtx)
	state := 0
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func(p int) {
			work := uint64(0)
			for benchmarking {
				work += 1
				mtx.Lock()
				for state%2 != p%2 {
					cv.Wait()
				}
				state += 1
				cv.Broadcast()
				mtx.Unlock()
			}
			compl <- work
		}(p)
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func SuiteCondvar(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"condvar-distr", benchmarkCondvarDistr})
	benchmarks.Push(&Benchmark{"condvar-nonblock", benchmarkCondvarNonblock})
	benchmarks.Push(&Benchmark{"condvar-block", func() {
		benchmarkCondvarBlock(1)
	}})
	benchmarks.Push(&Benchmark{"condvar-block-slack", func() {
		benchmarkCondvarBlock(8)
	}})
}
