// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

import (
	"fmt"
	"os"
	"flag"
	"regexp"
	"container/vector"
)

var flagDrawFile *string = flag.String("draw.file", "", "filename to draw (mandatory)")
var flagDrawBase *string = flag.String("draw.base", "", "base data filename for comparison (optional)")
var flagDrawRun *string = flag.String("draw.run", "", "regular expression to select benchmarks to draw (all by default)")
var flagDrawIdeal *bool = flag.Bool("draw.ideal", true, "do draw 'ideal' line (enabled by default)")

type BenchmarkRes struct {
	name    string
	base    bool
	through vector.IntVector
}

func parseFile(filename string) *vector.Vector {
	f, err := os.OpenFile(filename, os.O_RDONLY, 0)
	if f == nil {
		fmt.Printf("%s\n", err.String())
		os.Exit(1)
	}
	var benchmarks vector.Vector
	for true {
		var b BenchmarkRes
		n, _ := fmt.Fscanf(f, "%s", &b.name)
		if n == 0 {
			break
		}
		for true {
			var through int
			n, _ := fmt.Fscanf(f, "%d", &through)
			if n == 0 {
				break
			}
			b.through.Push(through)
		}
		benchmarks.Push(&b)
	}
	return &benchmarks
}

func createImageURL(benchmarks *vector.Vector) string {
	procs := 0
	maxThrough := 0
	maxThrough1 := maxThrough
	for bi := 0; bi != benchmarks.Len(); bi += 1 {
		b := benchmarks.At(bi).(*BenchmarkRes)
		if (procs != 0 && procs != b.through.Len()) || (b.through.Len() == 0) {
			fmt.Printf("proc count mismatch\n")
			os.Exit(1)
		}
		procs = b.through.Len()
		if maxThrough1 < b.through.At(0) {
			maxThrough1 = b.through.At(0)
		}
		for i := 0; i != b.through.Len(); i += 1 {
			if maxThrough < b.through.At(i) {
				maxThrough = b.through.At(i)
			}
		}
	}
	if *flagDrawIdeal {
		if maxThrough < maxThrough1*procs {
			maxThrough = maxThrough1 * procs
		}
	}

	// see http://code.google.com/apis/chart/image/docs/chart_params.html
	query := fmt.Sprintf("http://chart.googleapis.com/chart?cht=lc")

	// Image size
	query += "&chs=600x500"

	// Axis
	query += fmt.Sprintf("&chxt=x,y&chxr=0,1,%d,1%%7C1,0,%d", procs, maxThrough)

	// Margins
	query += "&chma=10,10,10,10"

	// Legend
	query += "&chdlp=b&chdl=ideal"
	for i := 0; i != benchmarks.Len(); i += 1 {
		b := benchmarks.At(i).(*BenchmarkRes)
		query += "%7C"
		query += b.name
		if b.base {
			query += "_base"
		}
	}

	// Grid
	query += fmt.Sprintf("&chg=%f,5", 100/float64(procs-1))

	// Colors
	query += fmt.Sprintf("&chco=AAAAAA")
	for i := 0; i != benchmarks.Len(); i += 1 {
		query += ","
		c := i
		if benchmarks.At(i).(*BenchmarkRes).base {
			c = c - benchmarks.Len()/2
		}
		if (c % 2) == 0 {
			query += "00"
		} else {
			query += "FF"
		}
		if ((c >> 1) % 2) == 0 {
			query += "00"
		} else {
			query += "FF"
		}
		if ((c >> 2) % 2) == 0 {
			query += "00"
		} else {
			query += "FF"
		}
	}

	// Style
	query += fmt.Sprintf("&chls=1")
	for i := 0; i != benchmarks.Len(); i += 1 {
		query += "%7C"
		if benchmarks.At(i).(*BenchmarkRes).base {
			query += "2,2,2"
		} else {
			query += "3"
		}
	}

	// Data series
	query += "&chd=t%3A"
	if *flagDrawIdeal {
		for p := 0; p != procs; p += 1 {
			if p != 0 {
				query += ","
			}
			query += fmt.Sprintf("%.2f", 100*float64(maxThrough1)*float64(p+1)/float64(maxThrough))
		}
	} else {
		query += "0"
	}
	for i := 0; i != benchmarks.Len(); i += 1 {
		b := benchmarks.At(i).(*BenchmarkRes)
		query += "%7C"
		for p := 0; p != procs; p += 1 {
			if p != 0 {
				query += ","
			}
			query += fmt.Sprintf("%.2f", 100*float64(b.through.At(p))/float64(maxThrough))
		}
	}

	return query
}

func doDraw() {
	benchmarks := parseFile(*flagDrawFile)
	if *flagDrawBase != "" {
		benchmarks0 := parseFile(*flagDrawBase)
		for i := 0; i != benchmarks0.Len(); i += 1 {
			b := benchmarks0.At(i).(*BenchmarkRes)
			b.base = true
			benchmarks.Push(b)
		}
	}
	for i := 0; i != benchmarks.Len(); i += 1 {
		b := benchmarks.At(i).(*BenchmarkRes)
		if match, _ := regexp.MatchString(*flagDrawRun, b.name); !match {
			benchmarks.Delete(i)
			i -= 1
		}
	}
	if benchmarks.Len() == 0 {
		fmt.Printf("no results found\n")
		os.Exit(1)
	}
	query := createImageURL(benchmarks)
	fmt.Printf("%s\n", query)
}
