// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests goroutine creation/completion performance/scalability.
// Spawns a lot of goroutines concurrently. goroutine-centr test uses single atomic
// variable to track when all the goroutines are created and executed (this single
// var can hinder scalability). goroutine-distr test uses distributed variables
// to track completion (it does not create scalability bottleneck, but puts additional
// pressure onto GC).
// Expected to scale almost linearly.

package main

import (
	"container/vector"
	"runtime"
	"sync/atomic"
)

const (
	LEVEL = 16
	COUNT = uint64(1<<uint(LEVEL+1) - 1)
)

type Compl struct {
	sync1 bool
	sync2 bool
}

func workerDistributed(level int, sync *bool) {
	if level > 0 {
		var compl Compl
		go workerDistributed(level-1, &compl.sync1)
		go workerDistributed(level-1, &compl.sync2)
		for compl.sync1 == false || compl.sync2 == false {
			runtime.Gosched()
		}
	}
	*sync = true
}

func benchmarkGorotuineDistributed() {
	for benchmarking {
		sync := false
		workerDistributed(LEVEL, &sync)
		totalWork += COUNT
	}
}

func workerCentralized(level int, sync *uint64) {
	if level > 0 {
		go workerCentralized(level-1, sync)
		go workerCentralized(level-1, sync)
	}
	atomic.AddUint64(sync, 1)
}

func benchmarkGorotuineCentralized() {
	for benchmarking {
		sync := uint64(0)
		workerCentralized(LEVEL, &sync)
		for sync != COUNT {
			runtime.Gosched()
		}
		totalWork += COUNT
	}
}

func SuiteGoroutine(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"goroutine-distr", benchmarkGorotuineDistributed})
	benchmarks.Push(&Benchmark{"goroutine-centr", benchmarkGorotuineCentralized})
}
