// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests scheduler scalability on matrix multiplication workload with fork/join style parallelism.
// There are 3 variations as to how to join child goroutines: by means of
// channels, by means of sync.WaitGroup or by means of direct shared memory notifications.
// There are also 3 grain sizes: 32, 16 and 8, that relate to leaf tasks that do
// 32x32, 16x16 and 8x8 matrix multiplication, respectively.
// Expected to scale almost linearly. 

package main

import (
	"container/vector"
	"runtime"
	"sync"
)

type Matrix [][]int

func makeMatrix(n int) Matrix {
	M := make(Matrix, n)
	for i := 0; i != n; i += 1 {
		M[i] = make([]int, n)
	}
	return M
}

func makeMatrixInit(n int) Matrix {
	M := makeMatrix(n)
	for i := 0; i != len(M); i += 1 {
		for j := 0; j != len(M[i]); j += 1 {
			M[i][j] = i*len(M) + j
		}
	}
	return M
}

// Parallel recursive cache-oblivious matrix multiplication.
// Synchronization is based on channels.
func matmultParallelImplChan(sync chan<- int, A, B, C Matrix, threshold, i0, i1, j0, j1, k0, k1 int) {
	di := i1 - i0
	dj := j1 - j0
	dk := k1 - k0
	if di >= dj && di >= dk && di >= threshold {
		mi := i0 + di/2
		sync0 := make(chan int, 1)
		go matmultParallelImplChan(sync0, A, B, C, threshold, i0, mi, j0, j1, k0, k1)
		matmultParallelImplChan(nil, A, B, C, threshold, mi, i1, j0, j1, k0, k1)
		<-sync0
	} else if dj >= dk && dj >= threshold {
		mj := j0 + dj/2
		sync0 := make(chan int, 1)
		go matmultParallelImplChan(sync0, A, B, C, threshold, i0, i1, j0, mj, k0, k1)
		matmultParallelImplChan(nil, A, B, C, threshold, i0, i1, mj, j1, k0, k1)
		<-sync0
	} else if dk >= threshold {
		mk := k0 + dk/2
		matmultParallelImplChan(nil, A, B, C, threshold, i0, i1, j0, j1, k0, mk)
		matmultParallelImplChan(nil, A, B, C, threshold, i0, i1, j0, j1, mk, k1)
	} else {
		for i := 0; i != 1000; i += 1 {
		}
		for i := i0; i < i1; i += 1 {
			for j := j0; j < j1; j += 1 {
				for k := k0; k < k1; k += 1 {
					C[i][j] += A[i][k] * B[k][j]
				}
			}
		}
	}
	if sync != nil {
		sync <- 0
	}
}

// Parallel recursive cache-oblivious matrix multiplication.
// Synchronization is based on *racy* reads and writes (formally it's incorrect).
func matmultParallelImplSync(sync *bool, A, B, C Matrix, threshold, i0, i1, j0, j1, k0, k1 int) {
	di := i1 - i0
	dj := j1 - j0
	dk := k1 - k0
	if di >= dj && di >= dk && di >= threshold {
		mi := i0 + di/2
		sync0 := false
		go matmultParallelImplSync(&sync0, A, B, C, threshold, i0, mi, j0, j1, k0, k1)
		matmultParallelImplSync(nil, A, B, C, threshold, mi, i1, j0, j1, k0, k1)
		for {
			if sync0 == true {
				break
			}
			runtime.Gosched()
		}
	} else if dj >= dk && dj >= threshold {
		mj := j0 + dj/2
		sync0 := false
		go matmultParallelImplSync(&sync0, A, B, C, threshold, i0, i1, j0, mj, k0, k1)
		matmultParallelImplSync(nil, A, B, C, threshold, i0, i1, mj, j1, k0, k1)
		for {
			if sync0 == true {
				break
			}
			runtime.Gosched()
		}
	} else if dk >= threshold {
		mk := k0 + dk/2
		matmultParallelImplSync(nil, A, B, C, threshold, i0, i1, j0, j1, k0, mk)
		matmultParallelImplSync(nil, A, B, C, threshold, i0, i1, j0, j1, mk, k1)
	} else {
		for i := i0; i < i1; i += 1 {
			for j := j0; j < j1; j += 1 {
				for k := k0; k < k1; k += 1 {
					C[i][j] += A[i][k] * B[k][j]
				}
			}
		}
	}
	if sync != nil {
		*sync = true
	}
}

// Parallel recursive cache-oblivious matrix multiplication.
// Synchronization is based on semaphores.
func matmultParallelImplWg(wg *sync.WaitGroup, A, B, C Matrix, threshold, i0, i1, j0, j1, k0, k1 int) {
	di := i1 - i0
	dj := j1 - j0
	dk := k1 - k0
	if di >= dj && di >= dk && di >= threshold {
		mi := i0 + di/2
		wg0 := &sync.WaitGroup{}
		wg0.Add(1)
		go matmultParallelImplWg(wg0, A, B, C, threshold, i0, mi, j0, j1, k0, k1)
		matmultParallelImplWg(nil, A, B, C, threshold, mi, i1, j0, j1, k0, k1)
		wg0.Wait()
	} else if dj >= dk && dj >= threshold {
		mj := j0 + dj/2
		wg0 := &sync.WaitGroup{}
		wg0.Add(1)
		go matmultParallelImplWg(wg0, A, B, C, threshold, i0, i1, j0, mj, k0, k1)
		matmultParallelImplWg(nil, A, B, C, threshold, i0, i1, mj, j1, k0, k1)
		wg0.Wait()
	} else if dk >= threshold {
		mk := k0 + dk/2
		matmultParallelImplWg(nil, A, B, C, threshold, i0, i1, j0, j1, k0, mk)
		matmultParallelImplWg(nil, A, B, C, threshold, i0, i1, j0, j1, mk, k1)
	} else {
		for i := i0; i < i1; i += 1 {
			for j := j0; j < j1; j += 1 {
				for k := k0; k < k1; k += 1 {
					C[i][j] += A[i][k] * B[k][j]
				}
			}
		}
	}
	if wg != nil {
		wg.Done()
	}
}

func benchmarkMatmult(impl func(A, B, C Matrix, threshold, n int), threshold int) {
	N := 128
	A := makeMatrixInit(N)
	B := makeMatrixInit(N)
	C := makeMatrix(N)
	for benchmarking {
		impl(A, B, C, threshold, N)
		totalWork += uint64(N * N * N)
	}
}

// Idiomatic channel-based
func benchmarkMatmultChan(threshold int) {
	impl := func(A, B, C Matrix, threshold, n int) {
		matmultParallelImplChan(nil, A, B, C, threshold, 0, n, 0, n, 0, n)
	}
	benchmarkMatmult(impl, threshold)
}

// Fast but incorrect
func benchmarkMatmultSync(threshold int) {
	impl := func(A, B, C Matrix, threshold, n int) {
		matmultParallelImplSync(nil, A, B, C, threshold, 0, n, 0, n, 0, n)
	}
	benchmarkMatmult(impl, threshold)
}

// Correct but slow (WaitGroup-based)
func benchmarkMatmultWg(threshold int) {
	impl := func(A, B, C Matrix, threshold, n int) {
		matmultParallelImplWg(nil, A, B, C, threshold, 0, n, 0, n, 0, n)
	}
	benchmarkMatmult(impl, threshold)
}

func SuiteMatmult(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"matmult-chan32", func() {
		benchmarkMatmultChan(32)
	}})
	benchmarks.Push(&Benchmark{"matmult-chan16", func() {
		benchmarkMatmultChan(16)
	}})
	benchmarks.Push(&Benchmark{"matmult-chan8", func() {
		benchmarkMatmultChan(8)
	}})

	benchmarks.Push(&Benchmark{"matmult-sync32", func() {
		benchmarkMatmultSync(32)
	}})
	benchmarks.Push(&Benchmark{"matmult-sync16", func() {
		benchmarkMatmultSync(16)
	}})
	benchmarks.Push(&Benchmark{"matmult-sync8", func() {
		benchmarkMatmultSync(8)
	}})

	benchmarks.Push(&Benchmark{"matmult-wg32", func() {
		benchmarkMatmultWg(32)
	}})
	benchmarks.Push(&Benchmark{"matmult-wg16", func() {
		benchmarkMatmultWg(16)
	}})
	benchmarks.Push(&Benchmark{"matmult-wg8", func() {
		benchmarkMatmultWg(8)
	}})
}
