// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests sync.Mutex performance/scalability.
// mutex-distr tests scalability of the mutex without contention.
//	Expected to scale almost linearly.
// mutex-centr tests scalability of the mutex under contention.
//	slack variations create more goroutines than there are worker threads.
//	work variations execute some amount of local work outside of critical section.
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
	"sync"
)

type PaddedMutex struct {
	mtx sync.Mutex
	pad [128]uint8
}

func benchmarkMutexDistr() {
	procs := runtime.GOMAXPROCS(-1)
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			var mtx PaddedMutex
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				mtx.mtx.Lock()
				mtx.mtx.Unlock()
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func benchmarkMutexCentr(slack, localWork int) {
	procs := runtime.GOMAXPROCS(-1) * slack
	compl := make(chan uint64, procs)
	var mtx sync.Mutex
	for p := 0; p != procs; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				mtx.Lock()
				mtx.Unlock()
				for i := 0; i != localWork; i += 1 {
					work *= 2
					work /= 2
				}
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func SuiteMutex(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"mutex-distr", func() {
		benchmarkMutexDistr()
	}})
	benchmarks.Push(&Benchmark{"mutex-centr", func() {
		benchmarkMutexCentr(1, 0)
	}})
	benchmarks.Push(&Benchmark{"mutex-centr-slack", func() {
		benchmarkMutexCentr(4, 0)
	}})
	benchmarks.Push(&Benchmark{"mutex-centr-work", func() {
		benchmarkMutexCentr(1, 100)
	}})
	benchmarks.Push(&Benchmark{"mutex-centr-slack-work", func() {
		benchmarkMutexCentr(4, 100)
	}})
}
