// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests sync.RWMutex performance/scalability.
// rwmutex-distr tests scalability of the mutex without contention.
//	Expected to scale almost linearly.
// rwmutex-w tests scalability of the mutex under contention.
//	w1/w10 variations refer to 1/10% write rate respectively.
//	slack variations create more goroutines than there are worker threads.
//	work variations execute some amount of local work in reader critical section.
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
	"sync"
)

type PaddedRWMutex struct {
	mtx sync.RWMutex
	pad [128]uint8
}

func benchmarkRWMutexDistr() {
	procs := runtime.GOMAXPROCS(-1)
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			mtx := &PaddedRWMutex{}
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				mtx.mtx.RLock()
				mtx.mtx.RUnlock()
				mtx.mtx.Lock()
				mtx.mtx.Unlock()
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func benchmarkRWMutexCentr(slack, wratio, localWork int) {
	procs := slack * runtime.GOMAXPROCS(-1)
	mtx := &sync.RWMutex{}
	compl := make(chan uint64, procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				if (work % uint64(wratio)) == 0 {
					mtx.Lock()
					mtx.Unlock()
				} else {
					mtx.RLock()
					for i := 0; i != localWork; i += 1 {
						work *= 2
						work /= 2
					}
					mtx.RUnlock()
				}
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func SuiteRWMutex(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"rwmutex-distr", benchmarkRWMutexDistr})
	benchmarks.Push(&Benchmark{"rwmutex-w1", func() {
		benchmarkRWMutexCentr(1, 100, 0)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w10", func() {
		benchmarkRWMutexCentr(1, 10, 0)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w1-slack", func() {
		benchmarkRWMutexCentr(4, 100, 0)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w10-slack", func() {
		benchmarkRWMutexCentr(4, 10, 0)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w1-work", func() {
		benchmarkRWMutexCentr(1, 100, 100)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w10-work", func() {
		benchmarkRWMutexCentr(1, 10, 100)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w1-slack-work", func() {
		benchmarkRWMutexCentr(4, 100, 100)
	}})
	benchmarks.Push(&Benchmark{"rwmutex-w10-slack-work", func() {
		benchmarkRWMutexCentr(4, 10, 100)
	}})
}
