// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests scalability of semaphores (runtime.Semacquire/Semrelease)
// sem-distr tests semaphore performance without contention.
// 	Each goroutine repeatedly releases and acquires a private semaphore.
//	Expected to scale almost linearly.
// Other tests examine semaphore performance under contention. 
//	nonblock variations stress fast-path, that is, no goroutines are blocked on the semaphore.
//	block variations test blocking slow-path, that is, some goroutines get blocked on the semaphore.
//	slack variations create more goroutines than there are worker threads (Ms).
//	work variations executine some amount of local work between semaphore operations.
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
)

type Padded struct {
	sem uint32
	pad [32]uint32
}

func benchmarkSem(slack, localWork int, block bool) {
	procs := runtime.GOMAXPROCS(-1) * slack
	compl := make(chan uint64, 2*procs)
	var sem uint32 = 0
	if block {
		for p := 0; p != procs/2; p += 1 {
			go func() {
				runtime.Semacquire(&sem)
				compl <- 0
			}()
		}
	}
	for p := 0; p != procs; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				runtime.Semrelease(&sem)
				for i := 0; i != localWork; i += 1 {
					work *= 2
					work /= 2
				}
				runtime.Semacquire(&sem)
			}
			compl <- work
		}()
	}
	if block {
		for p := 0; p != procs/2; p += 1 {
			totalWork += <-compl
		}
		for p := 0; p != procs; p += 1 {
			runtime.Semrelease(&sem)
			totalWork += <-compl
		}
	} else {
		for p := 0; p != procs; p += 1 {
			totalWork += <-compl
		}
	}
}

func benchmarkSemDistr() {
	procs := runtime.GOMAXPROCS(-1)
	compl := make(chan uint64, procs)
	for i := 0; i != procs; i += 1 {
		go func() {
			sem := &Padded{}
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				runtime.Semrelease(&sem.sem)
				runtime.Semacquire(&sem.sem)
			}
			compl <- work
		}()
	}
	for i := 0; i != procs; i += 1 {
		totalWork += <-compl
	}
}

func SuiteSem(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"sem-distr", benchmarkSemDistr})
	benchmarks.Push(&Benchmark{"sem-nonblock", func() {
		benchmarkSem(1, 0, false)
	}})
	benchmarks.Push(&Benchmark{"sem-nonblock-work", func() {
		benchmarkSem(1, 100, false)
	}})
	benchmarks.Push(&Benchmark{"sem-nonblock-slack", func() {
		benchmarkSem(4, 0, false)
	}})
	benchmarks.Push(&Benchmark{"sem-nonblock-slack-work", func() {
		benchmarkSem(4, 100, false)
	}})
	benchmarks.Push(&Benchmark{"sem-block", func() {
		benchmarkSem(1, 0, true)
	}})
	benchmarks.Push(&Benchmark{"sem-block-work", func() {
		benchmarkSem(1, 100, true)
	}})
	benchmarks.Push(&Benchmark{"sem-block-slack", func() {
		benchmarkSem(2, 0, true)
	}})
	benchmarks.Push(&Benchmark{"sem-block-slack-work", func() {
		benchmarkSem(2, 100, true)
	}})
}
